/*

  ECHOCLNT.C
  ==========

  Simple TCP/IP echo client using globus I/O with
  authentication and encryption included.

  Usage: echoclient <host_name> <port_number>

  Arto Ters <arto.teras@hip.fi> 18.6.2002

  Compare with a client using direct TCP socket calls, for example
  http://www.paulgriffiths.net/program/c/echoclnt.html

*/

/* Set AUTHENTICATE to one to use GSS-API authentication. */
#define AUTHENTICATE 1

/* Set ENCRYPTION to one to use ssl */
#define ENCRYPTION 1

#include <stdlib.h>
#include <stdio.h>

#include <globus_common.h>
#include <globus_io.h>
#include <globus_openssl.h>
#include <globus_gsi_credential.h>
#include <globus_gsi_proxy.h>

/*  Global constants  */
#define MAX_LINE            1000

int main(int argc, char *argv[]) {
    
    unsigned short int port;         /* Port number                     */
    char     *hostname;              /* Destination host name           */
    char     buffer[MAX_LINE];       /* Buffer to send data             */
    char     reply[MAX_LINE];        /* Buffer to receive data          */
    char     *endptr;                /* For strtol()                    */
    
    /* Variables for Globus functions */
    int                rval;         /* Return value (old APIs)         */
    globus_result_t    rptr;         /* Return value (new APIs)         */
    globus_object_t    *err;         /* Error type                      */
    globus_io_attr_t   attr;         /* I/O attributes of the server    */
    globus_io_handle_t conn_handle;  /* Globus socket handle            */
    globus_size_t      nbytes;       /* Number of bytes read or written */

    /* Variables for security functionality */
    globus_io_secure_authorization_data_t auth_data; /* Authentication data */

    /*  Get command line arguments  */
    if (argc != 3 ) {
	fprintf(stderr, "ECHOCLNT: Wrong number of arguments.\n");
	fprintf(stderr, "Usage: echoclnt <host_name> <port_number>\n");
	fprintf(stderr, "Example: echoclnt server.mydomain.com 1234\n");
	exit(EXIT_FAILURE);
    }
    else {
	hostname = argv[1];
	port = strtol(argv[2], &endptr, 0);
	if ( *endptr ) {
	    fprintf(stderr, "ECHOCLNT: Invalid port number.\n");
	    exit(EXIT_FAILURE);
	}
    }
    
    /* Activate the Globus I/O module */
    rval = globus_module_activate(GLOBUS_IO_MODULE);
    if (rval != GLOBUS_SUCCESS) {
	printf("ECHOCLNT: Globus I/O module activation failed\n");
	exit(EXIT_FAILURE);
    }
    rval = globus_module_activate(GLOBUS_GSI_GSS_ASSIST_MODULE);
    if (rval != GLOBUS_SUCCESS) {
	printf("ECHOCLNT: Globus I/O module activation failed\n");
	exit(EXIT_FAILURE);
    }
    rval = globus_module_activate(GLOBUS_GSI_CREDENTIAL_MODULE);
    if (rval != GLOBUS_SUCCESS) {
	printf("ECHOCLNT: Globus I/O module activation failed\n");
	exit(EXIT_FAILURE);
    }
    rval = globus_module_activate(GLOBUS_GSI_PROXY_MODULE);
    if (rval != GLOBUS_SUCCESS) {
	printf("ECHOCLNT: Globus I/O module activation failed\n");
	exit(EXIT_FAILURE);
    }
    rval = globus_module_activate(GLOBUS_GSI_CALLBACK_MODULE);
    if (rval != GLOBUS_SUCCESS) {
	printf("ECHOCLNT: Globus I/O module activation failed\n");
	exit(EXIT_FAILURE);
    }
    rval = globus_module_activate(GLOBUS_OPENSSL_MODULE);
    if (rval != GLOBUS_SUCCESS) {
	printf("ECHOCLNT: Globus I/O module activation failed\n");
	exit(EXIT_FAILURE);
    }

    /* Initialize the structure attr. */
    globus_io_tcpattr_init(&attr);

    if (AUTHENTICATE) {
	gss_cred_id_t *cred;
	gss_buffer_t buf;
	FILE *fp;
	char *file, *f2;
	struct stat s;
	OM_uint32 major;
	OM_uint32 minor;
	OM_uint32 timeRec;
	OM_uint32 time = 0;
	gss_OID p_desired_mech = NULL;
	int option_req = 0;

	cred = (gss_cred_id_t*)globus_malloc(sizeof(gss_cred_id_t));
	buf = (gss_buffer_desc*)globus_malloc(sizeof(gss_buffer_desc));

	f2 = "X509_USER_PROXY=/tmp/x509up_u424";

	file = "/tmp/pr";
	//file = "/tmp/x509up_u424";
	if (stat(file, &s) < 0)
	{
	    perror("stat");
	    exit(1);
	}
	
	buf->length = s.st_size;
	buf->value = (char *) globus_malloc(s.st_size);

	fp = fopen(file, "r");
	if (fp == NULL)
	{
	    perror("fopen");
	    exit(1);
	}

	if (fread(buf->value, buf->length, 1, fp) != 1)
	{
	    perror("fread");
	    exit(1);
	}
		
	/*
	  buf->value = f2;
	buf->length = strlen(f2);
	option_req = 1;
	*/

	major = gss_import_cred(&minor,
				cred,
				GSS_C_NO_OID,
				option_req,
				buf,
				(OM_uint32)time,
				&timeRec);
	if (major != GLOBUS_SUCCESS)
	{
	    printf("gss_import_cred failed\n");
	    exit(1);
	}


	/*
	 * Now  probe the cred we just loaded.
	 */

	{
	    gss_name_t name;
	    OM_uint32 lifetime;
	    gss_cred_usage_t usage;
	    gss_OID_set mechs;
	    
	    minor = 0;
	    major = gss_inquire_cred(&minor,
					    *cred,
					    &name,
					    &lifetime,
					    &usage,
					    &mechs);
	    if (major == GLOBUS_SUCCESS)
	    {
		gss_buffer_t gbuf;
		gss_OID           display_oid;
		
		gbuf = (gss_buffer_desc*)globus_malloc(sizeof(gss_buffer_desc));

		major = gss_display_name(&minor, name, gbuf, &display_oid);
		if (major == GLOBUS_SUCCESS)
		{
		    printf("Name: %s\n", gbuf->value);
		}
	    }
	}

	/* Use GSS-API authentication */
	rptr = globus_io_attr_set_secure_authentication_mode(&attr, 
             GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI, *cred);
	if (rptr != GLOBUS_SUCCESS) {
	    printf("ECHOCLNT: Setting secure authentication mode failed\n");
	    exit(EXIT_FAILURE);
	}

	/* Initialize the authorization data */
	rptr = globus_io_secure_authorization_data_initialize(&auth_data);
	if (rptr != GLOBUS_SUCCESS) {
	    printf("ECHOCLNT: Initializing authorization data failed\n");
	    exit(EXIT_FAILURE);
	}
 
	/* Set the authorization mode */
	rptr = globus_io_attr_set_secure_authorization_mode(&attr, 
             GLOBUS_IO_SECURE_AUTHORIZATION_MODE_SELF, &auth_data);
	if (rptr != GLOBUS_SUCCESS) {
	    printf("ECHOCLNT: Setting authorization mode failed\n");
	    exit(EXIT_FAILURE);
	}

	if (ENCRYPTION) {
	    /* Set the channel mode to use wrapping */
	    rptr = globus_io_attr_set_secure_channel_mode(&attr, 
                 GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP);
	    if (rptr != GLOBUS_SUCCESS) {
		printf("ECHOCLNT: Setting channel mode failed\n");
		exit(EXIT_FAILURE);
	    }

	    /* Use encryption, SAFE instead of PRIVATE is also possible
	       but gives only integrity protection */
	    rptr = globus_io_attr_set_secure_protection_mode(&attr, 
                 GLOBUS_IO_SECURE_PROTECTION_MODE_PRIVATE);
	    if (rptr != GLOBUS_SUCCESS) {
		printf("ECHOCLNT: Setting encrypted mode failed\n");
		exit(EXIT_FAILURE);
	    }   
	}
    }
    
    /*  Connect to the remote echo server  */
    rptr = globus_io_tcp_connect(hostname, port, &attr, &conn_handle);
    if (rptr != GLOBUS_SUCCESS) {
	/* Expand the error object and report the type of failure */
	err = globus_error_get(rptr);
	if (globus_object_type_match(GLOBUS_IO_ERROR_TYPE_AUTHENTICATION_FAILED, 
				     globus_object_get_type(err))) {
	    globus_libc_printf("ECHOCLNT: Authentication failed\n");
	}
	else if (globus_object_type_match(GLOBUS_IO_ERROR_TYPE_AUTHORIZATION_FAILED,
					  globus_object_get_type(err))) {
	    globus_libc_printf("ECHOCLNT: Authorization failed\n");
	}
	else if (globus_object_type_match(GLOBUS_IO_ERROR_TYPE_NO_CREDENTIALS,
					  globus_object_get_type(err))) {
	    globus_libc_printf("ECHOCLNT: Failure: no credentials\n");
	}
	else {
	    globus_libc_printf("ECHOCLNT: Connecting to the remote server failed\n");
	}
	
	globus_object_free(err);	
	exit(EXIT_FAILURE);
    }

    /*  Get string to echo from user  */
    printf("Enter the string to echo: ");
    fgets(buffer, MAX_LINE, stdin);
    
    /*  Send string to echo server, and retrieve response  */
    globus_io_write(&conn_handle, buffer, strlen(buffer), &nbytes);
    globus_io_read(&conn_handle, reply, MAX_LINE-1, 1, &nbytes);
    reply[nbytes]='\0';

    /*  Output echoed string  */
    printf("Echo response: %s\n", reply);

    /* Close the socket */
    rptr = globus_io_close(&conn_handle);
    if (rptr != GLOBUS_SUCCESS) {
	printf("ECHOCLNT: Closing the socket failed\n");
 	exit(EXIT_FAILURE);
    }

    /* Deactivate the Globus I/O module */
    rval = globus_module_deactivate(GLOBUS_IO_MODULE);
    if (rval != GLOBUS_SUCCESS) {
	printf("ECHOCLNT: Globus I/O module deactivation failed\n");
	exit(EXIT_FAILURE);
    }

    return EXIT_SUCCESS;
}

